from collections import namedtuple
from functools import wraps
from typing import Dict
from typing import Iterable
from typing import Optional

from flask import request
from flask import session

from opwen_email_client.domain.email.attachment import AttachmentEncoder
from opwen_email_client.domain.email.store import EmailStore


# noinspection PyClassHasNoInit
class FileInfo(namedtuple('FileInfo', 'name content')):
    pass


class AttachmentsStore(object):
    def __init__(self, attachment_encoder: AttachmentEncoder,
                 email_store: EmailStore):
        self._attachment_encoder = attachment_encoder
        self._email_store = email_store

    @property
    def _session_store(self) -> dict:
        session_attachments = session.get('attachments')
        if not session_attachments:
            session_attachments = session['attachments'] = {}

        return session_attachments

    def store(self, emails: Iterable[dict]):
        for i, email in enumerate(emails):
            email_id = email['_uid']
            attachments = email.get('attachments', [])
            for j, attachment in enumerate(attachments):
                attachment_filename = attachment.get('filename', '')
                attachment_id = '{}-{}'.format(i + 1, attachment_filename)
                attachment['id'] = attachment_id
                self._session_store[attachment_id] = (email_id, j)

    def lookup(self, attachment_id: str) -> Optional[FileInfo]:
        try:
            email_id, attachment_idx = self._session_store[attachment_id]
        except KeyError:
            return None

        email = self._email_store.get(email_id)
        if email is None:
            return None

        attachments = email.get('attachments', [])
        if attachment_idx >= len(attachments):
            return None

        attachment = attachments[attachment_idx]  # type: Dict
        filename = attachment.get('filename')
        content = attachment.get('content')
        if not filename or not content:
            return None

        return FileInfo(filename, self._attachment_encoder.decode(content))


class Session(object):
    _current_language_key = 'current_language'
    _last_visited_url_key = 'last_visited_url'

    @classmethod
    def _session(cls) -> Dict[str, str]:
        return session

    @classmethod
    def store_last_visited_url(cls):
        cls._session()[cls._last_visited_url_key] = request.url

    @classmethod
    def get_last_visited_url(cls) -> Optional[str]:
        return cls._session().get(cls._last_visited_url_key)

    @classmethod
    def store_current_language(cls, language: str):
        cls._session()[cls._current_language_key] = language

    @classmethod
    def get_current_language(cls) -> str:
        return cls._session().get(cls._current_language_key)


def track_history(func):
    @wraps(func)
    def history_tracker(*args, **kwargs):
        Session.store_last_visited_url()
        return func(*args, **kwargs)

    return history_tracker



from datetime import datetime

from sqlalchemy import Boolean
from sqlalchemy import Column
from sqlalchemy import DateTime
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy import Table
from sqlalchemy import Text
from sqlalchemy import exists
from sqlalchemy import or_
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship
from sqlalchemy.orm import sessionmaker

from opwen_email_client.domain.email.store import EmailStore
from opwen_email_client.util.sqlalchemy import create_database
from opwen_email_client.util.sqlalchemy import get_or_create
from opwen_email_client.util.sqlalchemy import session

_Base = declarative_base()


_EmailTo = Table('emailto',
                 _Base.metadata,
                 Column('email_id', Integer, ForeignKey('email.id')),
                 Column('to_id', Integer, ForeignKey('to.id')))

_EmailCc = Table('emailcc',
                 _Base.metadata,
                 Column('email_id', Integer, ForeignKey('email.id')),
                 Column('cc_id', Integer, ForeignKey('cc.id')))

_EmailBcc = Table('emailbcc',
                  _Base.metadata,
                  Column('email_id', Integer, ForeignKey('email.id')),
                  Column('bcc_id', Integer, ForeignKey('bcc.id')))

_EmailAttachment = Table(
    'emailattachment',
    _Base.metadata,
    Column('email_id', Integer, ForeignKey('email.id')),
    Column('attachment_id', Integer, ForeignKey('attachment.id')))


class _To(_Base):
    __tablename__ = 'to'
    id = Column(Integer, primary_key=True)

    address = Column(String(length=128), index=True, unique=True)


class _Cc(_Base):
    __tablename__ = 'cc'
    id = Column(Integer, primary_key=True)

    address = Column(String(length=128), index=True, unique=True)


class _Bcc(_Base):
    __tablename__ = 'bcc'
    id = Column(Integer, primary_key=True)

    address = Column(String(length=128), index=True, unique=True)


class _Attachment(_Base):
    __tablename__ = 'attachment'
    id = Column(Integer, primary_key=True)

    filename = Column(Text)
    content = Column(Text)


class _Email(_Base):
    __tablename__ = 'email'
    id = Column(Integer, primary_key=True)

    uid = Column(String(length=64), unique=True, index=True)
    subject = Column(Text)
    body = Column(Text)
    sent_at = Column(DateTime())
    read = Column(Boolean, default=False, nullable=False)
    sender = Column(String(length=128), index=True)
    attachments = relationship(_Attachment, secondary=_EmailAttachment,
                               backref='emails')
    to = relationship(_To, secondary=_EmailTo)
    cc = relationship(_Cc, secondary=_EmailCc)
    bcc = relationship(_Bcc, secondary=_EmailBcc)

    def to_dict(self):
        attachments = self.attachments
        attachments = ([{'filename': attachment.filename,
                         'content': attachment.content}
                        for attachment in attachments]
                       if attachments else None)

        sent_at = self.sent_at
        sent_at = (sent_at.strftime('%Y-%m-%d %H:%M')
                   if sent_at else None)

        return {k: v for (k, v) in (
            ('from', self.sender),
            ('to', [_.address for _ in self.to]),
            ('cc', [_.address for _ in self.cc]),
            ('bcc', [_.address for _ in self.bcc]),
            ('subject', self.subject),
            ('body', self.body),
            ('_uid', self.uid),
            ('sent_at', sent_at),
            ('read', self.read),
            ('attachments', attachments),
        ) if v}

    @classmethod
    def from_dict(cls, db, email):
        sent_at = email.get('sent_at')
        sent_at = (datetime.strptime(sent_at, '%Y-%m-%d %H:%M')
                   if sent_at else None)

        return _Email(
            uid=email['_uid'],
            to=[get_or_create(db, _To, address=_.lower())
                for _ in email.get('to', [])],
            cc=[get_or_create(db, _Cc, address=_.lower())
                for _ in email.get('cc', [])],
            bcc=[get_or_create(db, _Bcc, address=_.lower())
                 for _ in email.get('bcc', [])],
            attachments=[get_or_create(db, _Attachment, **_)
                         for _ in email.get('attachments', [])],
            subject=email.get('subject'),
            body=email.get('body'),
            sent_at=sent_at,
            read=email.get('read', False),
            sender=email.get('from', '').lower() or None)

    @classmethod
    def is_sent_by(cls, email_address):
        email_address = email_address.lower()
        return cls.sender == email_address

    @classmethod
    def is_received_by(cls, email_address):
        email_address = email_address.lower()
        return (cls.to.any(_To.address == email_address) |
                cls.cc.any(_Cc.address == email_address) |
                cls.bcc.any(_Bcc.address == email_address))


class _SqlalchemyEmailStore(EmailStore):
    def __init__(self, database_uri: str):
        self._base = _Base
        self._engine = create_database(database_uri, self._base)
        self._sesion_maker = sessionmaker(autocommit=False, autoflush=False,
                                          bind=self._engine)

    def _dbread(self):
        return session(self._sesion_maker, commit=False)

    def _dbwrite(self):
        return session(self._sesion_maker, commit=True)

    def _create(self, emails):
        with self._dbwrite() as db:
            for email in emails:
                uid_exists = exists().where(_Email.uid == email['_uid'])
                if not db.query(uid_exists).scalar():
                    db.add(_Email.from_dict(db, email))

    def _mark_sent(self, uids):
        now = datetime.utcnow()
        set_sent_at = {_Email.sent_at: now}

        with self._dbwrite() as db:
            db.query(_Email)\
                .filter(_match_email_uid(uids))\
                .update(set_sent_at)

    def _mark_read(self, email_address, uids):
        set_read = {_Email.read: True}

        with self._dbwrite() as db:
            db.query(_Email)\
                .filter(_match_email_uid(uids) & _can_access(email_address))\
                .update(set_read, synchronize_session='fetch')

    def _delete(self, email_address, uids):
        should_delete = _match_email_uid(uids) & _can_access(email_address)

        with self._dbwrite() as db:
            for email in db.query(_Email).filter(should_delete).all():
                email.attachments = []
                db.delete(email)

        self._delete_orphaned_attachments()

    def _delete_orphaned_attachments(self):
        with self._dbwrite() as db:
            db.query(_Attachment)\
                .filter(~_Attachment.emails.any())\
                .delete(synchronize_session='fetch')

    def _find(self, query):
        with self._dbread() as db:
            results = db.query(_Email).filter(query)
            email = results.first()
            return email.to_dict() if email else None

    def _query(self, query):
        with self._dbread() as db:
            results = db.query(_Email).filter(query)
            for email in results.order_by(_Email.sent_at.desc()).all():
                yield email.to_dict()

    def inbox(self, email_address):
        return self._query(_Email.is_received_by(email_address))

    def outbox(self, email_address):
        return self._query(_Email.is_sent_by(email_address)
                           & _Email.sent_at.is_(None))

    def search(self, email_address, query):
        textquery = '%{}%'.format(query)
        contains_query = or_(*(_Email.subject.ilike(textquery),
                               _Email.body.ilike(textquery),
                               _Email.sender.ilike(textquery),
                               _Email.to.any(_To.address.ilike(textquery)),
                               _Email.cc.any(_Cc.address.ilike(textquery)),
                               _Email.bcc.any(_Bcc.address.ilike(textquery))))
        return self._query(_can_access(email_address) & contains_query)

    def pending(self):
        return self._query(_Email.sent_at.is_(None))

    def get(self, uid):
        return self._find(_Email.uid == uid)

    def sent(self, email_address):
        return self._query(_Email.is_sent_by(email_address)
                           & _Email.sent_at.isnot(None))


class SqliteEmailStore(_SqlalchemyEmailStore):
    def __init__(self, database_path: str):
        super().__init__('sqlite:///{}'.format(database_path))


def _can_access(email_address):
    return (_Email.is_sent_by(email_address)
            | _Email.is_received_by(email_address))


def _match_email_uid(uids):
    return or_(*(_Email.uid == uid for uid in uids))



from abc import ABCMeta
from abc import abstractmethod
from gzip import GzipFile
from io import BytesIO
from io import TextIOBase
from os import getenv
from os import makedirs
from os import path
from shutil import copyfileobj
from tempfile import NamedTemporaryFile
from typing import Iterable
from typing import TypeVar
from uuid import uuid4

from azure.common import AzureMissingResourceHttpError
from azure.storage.blob import BlockBlobService

from opwen_email_client.domain.email.client import EmailServerClient
from opwen_email_client.util.serialization import Serializer

T = TypeVar('T')


class Sync(metaclass=ABCMeta):
    @abstractmethod
    def upload(self, items: Iterable[T]) -> Iterable[str]:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def download(self) -> Iterable[T]:
        raise NotImplementedError  # pragma: no cover


class AzureSync(Sync):
    def __init__(self, container: str, serializer: Serializer,
                 account_name: str, account_key: str,
                 email_server_client: EmailServerClient,
                 azure_client: BlockBlobService=None):

        self._container = container
        self._serializer = serializer
        self._account_name = account_name
        self._account_key = account_key
        self._email_server_client = email_server_client
        self.__azure_client = azure_client

    @property
    def _azure_client(self) -> BlockBlobService:
        if not self.__azure_client:
            self.__azure_client = BlockBlobService(self._account_name,
                                                   self._account_key)
        return self.__azure_client

    @classmethod
    def _workspace(cls) -> TextIOBase:
        return NamedTemporaryFile()

    @classmethod
    def _open(cls, fileobj: BytesIO, mode: str='rb') -> TextIOBase:
        return GzipFile(fileobj=fileobj, mode=mode)

    def _download_to_stream(self, blobname: str, container: str,
                            stream: TextIOBase) -> bool:

        try:
            self._azure_client.get_blob_to_stream(container, blobname, stream)
        except AzureMissingResourceHttpError:
            return False
        else:
            return True

    def _upload_from_stream(self, blobname: str, stream: TextIOBase):
        self._azure_client.create_blob_from_stream(self._container,
                                                   blobname, stream)

    def download(self):
        resource_id, container = self._email_server_client.download()
        if not resource_id or not container:
            return

        with self._workspace() as workspace:
            if self._download_to_stream(resource_id, container, workspace):
                workspace.seek(0)
                with self._open(workspace) as downloaded:
                    for line in downloaded:
                        yield self._serializer.deserialize(line)

    def upload(self, items):
        uploaded_ids = []
        upload_location = str(uuid4())

        with self._workspace() as workspace:
            with self._open(workspace, 'wb') as uploaded:
                for item in items:
                    item = {key: value for (key, value) in item.items()
                            if value is not None}
                    serialized = self._serializer.serialize(item)
                    uploaded.write(serialized)
                    uploaded.write(b'\n')
                    uploaded_ids.append(item.get('_uid'))

            if uploaded_ids:
                workspace.seek(0)
                self._upload_from_stream(upload_location, workspace)
                self._email_server_client.upload(upload_location,
                                                 self._container)

        return uploaded_ids


class LocalAzureSync(AzureSync):
    def _upload_from_stream(self, blobname: str, stream: TextIOBase):
        upload_directory = path.join(getenv('AZURE_ROOT'), self._container)
        makedirs(upload_directory, exist_ok=True)
        local_filename = path.join(upload_directory, blobname)

        with open(local_filename, 'wb') as fobj:
            copyfileobj(stream, fobj)

    def _download_to_stream(self, blobname: str, container: str,
                            stream: TextIOBase):
        download_directory = path.join(getenv('AZURE_ROOT'), container)
        local_filename = path.join(download_directory, blobname)

        try:
            with open(local_filename, 'rb') as fobj:
                copyfileobj(fobj, stream)
        except FileNotFoundError:
            return False
        else:
            return True



#!/usr/bin/env python3

from flask_migrate import MigrateCommand
from flask_script import Manager

from opwen_email_client.util.management import DevServerCommand
from opwen_email_client.webapp import app

manager = Manager(app)
manager.add_command('db', MigrateCommand)
manager.add_command('devserver', DevServerCommand)

manager.run()



from logging import ERROR
from os import path
from tempfile import gettempdir

from babel import Locale
from flask_babel import gettext as _

from opwen_email_client.util.os import getenv
from opwen_email_client.util.os import subdirectories


state_basedir = path.abspath(getenv('OPWEN_STATE_DIRECTORY', gettempdir()))
app_basedir = path.abspath(path.dirname(__file__))


# noinspection PyPep8Naming
class i8n(object):
    LOGIN_REQUIRED = _('Please log in to access this page.')
    UNAUTHORIZED = _('You do not have permission to view this page.')
    INVALID_PASSWORD = _('Invalid password.')
    SHORT_PASSWORD = _('The password must have at least 6 characters.')
    EMAIL_SENT = _('Email sent!')
    EMAIL_ADDRESS_INVALID = _('Invalid email address.')
    EMAIL_TO_REQUIRED = _('Please specify a recipient.')
    EMAIL_NO_SUBJECT = _('(no subject)')
    LOGGED_IN = _('You are now logged in.')
    LOGGED_OUT = _('You have logged out successfully.')
    WELCOME = _('Welcome!')
    EMAIL_CHARACTERS = _('Please only use letters, numbers, dots and dashes.')
    ACCOUNT_CREATED = _('Your Lokole account has been created successfully!')
    ACCOUNT_SUSPENDED = _('Your account has been suspended. '
                          'Please contact your administrator.')
    SYNC_COMPLETE = _('Email synchronization completed.')
    UNEXPECTED_ERROR = _('Unexpected error. Please contact your admin.')
    PAGE_DOES_NOT_EXIST = _('This page does not exist.')
    USER_DOES_NOT_EXIST = _('This user does not exist.')
    USER_SUSPENDED = _('The user was suspended.')
    USER_UNSUSPENDED = _('The user was un-suspended.')
    ADMIN_CANNOT_BE_SUSPENDED = _("Administrators can't be suspended.")
    PASSWORD_CHANGED_BY_ADMIN = _('Password was reset by administrator to: ')
    SAME_PASSWORD = _(' Your new password must be different than your '
                      'previous password.')


class AppConfig(object):
    SQLITE_PATH = path.join(state_basedir, 'users.sqlite3')
    SQLALCHEMY_DATABASE_URI = 'sqlite:///' + SQLITE_PATH
    SQLALCHEMY_MIGRATE_REPO = path.join(state_basedir, 'app.migrate')
    SQLALCHEMY_TRACK_MODIFICATIONS = False

    ADMIN_SECRET = getenv('OPWEN_ADMIN_SECRET')
    SECRET_KEY = getenv('OPWEN_SESSION_KEY')
    SECURITY_USER_IDENTITY_ATTRIBUTES = 'email'
    SECURITY_PASSWORD_HASH = 'bcrypt'
    SECURITY_PASSWORD_SALT = getenv('OPWEN_PASSWORD_SALT')
    SECURITY_REGISTERABLE = True
    SECURITY_CHANGEABLE = True
    SECURITY_TRACKABLE = True
    SECURITY_SEND_PASSWORD_CHANGE_EMAIL = False
    SECURITY_POST_REGISTER_VIEW = 'register_complete'
    SECURITY_POST_LOGIN_VIEW = 'login_complete'
    SECURITY_POST_LOGOUT_VIEW = 'logout_complete'
    SECURITY_SEND_REGISTER_EMAIL = False
    SECURITY_MSG_LOGIN = i8n.LOGIN_REQUIRED, 'error'
    SECURITY_MSG_UNAUTHORIZED = i8n.UNAUTHORIZED, 'error'
    SECURITY_MSG_INVALID_PASSWORD = i8n.INVALID_PASSWORD, 'error'
    SECURITY_MSG_DISABLED_ACCOUNT = i8n.ACCOUNT_SUSPENDED, 'error'
    SECURITY_MSG_PASSWORD_INVALID_LENGTH = i8n.SHORT_PASSWORD, 'error'
    SECURITY_MSG_PASSWORD_IS_THE_SAME = i8n.SAME_PASSWORD, 'error'
    SECURITY_LOGIN_URL = '/user/login'
    SECURITY_LOGOUT_URL = '/user/logout'
    SECURITY_REGISTER_URL = '/user/register'
    SECURITY_CHANGE_URL = '/user/password/change'

    TESTING = getenv('OPWEN_ENABLE_DEBUG', False)

    LOCAL_EMAIL_STORE = path.join(state_basedir, 'emails.sqlite3')

    EMAIL_ADDRESS_DELIMITER = ','
    EMAILS_PER_PAGE = 30

    LOG_FORMAT = '%(asctime)s\t%(levelname)s\t%(message)s'
    LOG_LEVEL = ERROR

    LOCALES_DIRECTORY = path.join(app_basedir, 'translations')
    DEFAULT_LOCALE = Locale.parse('en_ca')
    LOCALES = (
        [DEFAULT_LOCALE] +
        [Locale.parse(code) for code in subdirectories(LOCALES_DIRECTORY)])

    EMAIL_SERVER_READ_API_HOSTNAME = getenv('OPWEN_EMAIL_SERVER_READ_API')
    EMAIL_SERVER_WRITE_API_HOSTNAME = getenv('OPWEN_EMAIL_SERVER_WRITE_API')
    EMAIL_HOST_FORMAT = '{}.lokole.ca'
    STORAGE_CONTAINER = 'compressedpackages'
    STORAGE_ACCOUNT_NAME = getenv('OPWEN_REMOTE_ACCOUNT_NAME')
    STORAGE_ACCOUNT_KEY = getenv('OPWEN_REMOTE_ACCOUNT_KEY')
    CLIENT_NAME = getenv('OPWEN_CLIENT_NAME')
    CLIENT_ID = getenv('OPWEN_CLIENT_ID')
    CLIENT_EMAIL_HOST = EMAIL_HOST_FORMAT.format(CLIENT_NAME)



from logging import Formatter
from logging import StreamHandler

from flask import Flask
from flask_babel import Babel

from opwen_email_client.domain.email.attachment import Base64AttachmentEncoder as AttachmentEncoder  # noqa
from opwen_email_client.domain.email.client import HttpEmailServerClient as EmailServerClient  # noqa
from opwen_email_client.domain.email.sql_store import SqliteEmailStore as EmailStore  # noqa
from opwen_email_client.domain.email.sync import AzureSync as Sync  # noqa
from opwen_email_client.util.serialization import JsonSerializer as Serializer  # noqa
from opwen_email_client.webapp.config import AppConfig
from opwen_email_client.webapp.session import AttachmentsStore


if AppConfig.TESTING:
    from opwen_email_client.domain.email.client import LocalEmailServerClient as EmailServerClient  # noqa
    from opwen_email_client.domain.email.sync import LocalAzureSync as Sync  # noqa


class Ioc(object):
    serializer = Serializer()

    email_server_client = EmailServerClient(
        read_api=AppConfig.EMAIL_SERVER_READ_API_HOSTNAME,
        write_api=AppConfig.EMAIL_SERVER_WRITE_API_HOSTNAME,
        client_id=AppConfig.CLIENT_ID)

    email_store = EmailStore(
        database_path=AppConfig.LOCAL_EMAIL_STORE)

    email_sync = Sync(
        account_name=AppConfig.STORAGE_ACCOUNT_NAME,
        account_key=AppConfig.STORAGE_ACCOUNT_KEY,
        email_server_client=email_server_client,
        container=AppConfig.STORAGE_CONTAINER,
        serializer=serializer)

    attachment_encoder = AttachmentEncoder()

    attachments_session = AttachmentsStore(
        email_store=email_store,
        attachment_encoder=attachment_encoder)


def create_app() -> Flask:
    app = Flask(__name__)
    app.config.from_object(AppConfig)
    app.ioc = Ioc()

    handler = StreamHandler()
    handler.setFormatter(Formatter(AppConfig.LOG_FORMAT))
    app.logger.addHandler(handler)
    app.logger.setLevel(AppConfig.LOG_LEVEL)

    app.babel = Babel(app)

    return app



from abc import ABCMeta
from abc import abstractmethod
from typing import Iterable
from typing import Optional
from typing import Union
from uuid import uuid4


class EmailStore(metaclass=ABCMeta):
    def create(self, emails: Iterable[dict]):
        self._create(map(_add_uid, emails))

    @abstractmethod
    def _create(self, emails: Iterable[dict]):
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def get(self, uid: str) -> Optional[dict]:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def inbox(self, email_address: str) -> Iterable[dict]:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def outbox(self, email_address: str) -> Iterable[dict]:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def sent(self, email_address: str) -> Iterable[dict]:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def search(self, email_address: str,
               query: Optional[str]) -> Iterable[dict]:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def pending(self) -> Iterable[dict]:
        raise NotImplementedError  # pragma: no cover

    def mark_sent(self, emails_or_uids: Iterable[Union[dict, str]]):
        uids = map(_get_uid, emails_or_uids)
        return self._mark_sent(uids)

    def mark_read(self, email_address: str,
                  emails_or_uids: Iterable[Union[dict, str]]):
        uids = map(_get_uid, emails_or_uids)
        return self._mark_read(email_address, uids)

    def delete(self, email_address: str,
               emails_or_uids: Iterable[Union[dict, str]]):
        uids = map(_get_uid, emails_or_uids)
        return self._delete(email_address, uids)

    @abstractmethod
    def _delete(self, email_address: str, uids: Iterable[str]):
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def _mark_sent(self, uids: Iterable[str]):
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def _mark_read(self, email_address: str, uids: Iterable[str]):
        raise NotImplementedError  # pragma: no cover


def _get_uid(email_or_uid: Union[str, dict]) -> str:
    try:
        return email_or_uid['_uid']
    except TypeError:
        return email_or_uid


def _add_uid(email: dict) -> dict:
    email.setdefault('_uid', str(uuid4()))
    return email



try:
    from html import unescape
except ImportError:
    from html.parser import HTMLParser
    unescape = HTMLParser().unescape  # type: ignore

from re import IGNORECASE
from re import compile as re_compile
from typing import Optional

from wtforms import StringField
from wtforms import TextAreaField
from wtforms.validators import HostnameValidation
from wtforms.validators import Regexp
from wtforms.validators import ValidationError


class Emails(Regexp):
    def __init__(self, email_address_delimiter, message=None):
        self.validate_hostname = HostnameValidation(require_tld=True)
        self.email_address_delimiter = email_address_delimiter
        super().__init__(r'^.+@([^.@][^@]+)$', IGNORECASE, message)

    # noinspection PyMethodOverriding
    def __call__(self, form, field):
        message = self.message
        if message is None:
            message = field.gettext('Invalid email address.')

        messages = message.split(self.email_address_delimiter)
        for message in messages:
            match = super().__call__(form, field, message.strip())
            if not self.validate_hostname(match.group(1)):
                raise ValidationError(message)


class HtmlTextAreaField(TextAreaField):
    _script_tags_re = re_compile(r'<script[\s\S]+?/script>', IGNORECASE)

    # noinspection PyAttributeOutsideInit
    def process_formdata(self, valuelist):
        super().process_formdata(valuelist)
        self.data = self._to_safe_html(self.data)

    def _value(self):
        value = super()._value()
        return self._to_safe_html(value)

    @classmethod
    def _remove_script_tags(cls, data: str) -> str:
        match = cls._script_tags_re.search(data)
        while match:
            start = match.start()
            end = match.end()
            data = data[:start] + data[end:]
            match = cls._script_tags_re.search(data)

        return data

    @classmethod
    def _to_safe_html(cls, data: Optional[str]) -> str:
        if not data:
            return ''

        data = unescape(data)
        data = cls._remove_script_tags(data)
        return data


class SuffixedStringField(StringField):
    def __init__(self, suffix: str='', *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._suffix = suffix

    # noinspection PyAttributeOutsideInit
    def process_formdata(self, valuelist):
        super().process_formdata(valuelist)
        self.data = self._ensure_suffix(self.data)

    def _value(self):
        value = super()._value()
        return self._ensure_suffix(value)

    def _ensure_suffix(self, value: str) -> str:
        if value and not value.endswith(self._suffix):
            return value + self._suffix
        else:
            return value



from glob import glob
from os.path import join
from typing import List

from flask import Flask
from flask_script import Command


# noinspection PyAbstractClass,PyMethodOverriding
class DevServerCommand(Command):
    def __call__(self, app: Flask):
        _load_environment(app)

        templates_paths = _templates_paths_for(app, '*.html')
        reload_server_if_changed = templates_paths

        app.run(debug=True, extra_files=reload_server_if_changed)


def _load_environment(app: Flask) -> None:
    try:
        # noinspection PyUnresolvedReferences
        from dotenv import load_dotenv
        load_dotenv(join(app.root_path, '..', '..', '.env'))
    except ImportError:
        pass


def _templates_paths_for(app: Flask, templates_matcher: str) -> List[str]:
    templates_directory = join(app.root_path, app.template_folder)
    templates_glob = join(templates_directory, '**', templates_matcher)
    return glob(templates_glob, recursive=True)



from datetime import datetime
from itertools import chain
from typing import Iterable
from typing import List
from typing import Optional

from flask import render_template
from flask import request
from flask_login import current_user
from flask_wtf import Form
from werkzeug.datastructures import FileStorage
from wtforms import FileField
from wtforms import StringField
from wtforms import SubmitField
from wtforms.validators import DataRequired
from wtforms.validators import Optional as DataOptional

from opwen_email_client.domain.email.attachment import AttachmentEncoder
from opwen_email_client.domain.email.store import EmailStore
from opwen_email_client.util.wtforms import Emails
from opwen_email_client.util.wtforms import HtmlTextAreaField
from opwen_email_client.webapp.config import AppConfig
from opwen_email_client.webapp.config import i8n


class NewEmailForm(Form):
    to = StringField(
        validators=[DataRequired(i8n.EMAIL_TO_REQUIRED),
                    Emails(AppConfig.EMAIL_ADDRESS_DELIMITER,
                           i8n.EMAIL_ADDRESS_INVALID)])

    cc = StringField(
        validators=[DataOptional(),
                    Emails(AppConfig.EMAIL_ADDRESS_DELIMITER,
                           i8n.EMAIL_ADDRESS_INVALID)])

    bcc = StringField(
        validators=[DataOptional(),
                    Emails(AppConfig.EMAIL_ADDRESS_DELIMITER,
                           i8n.EMAIL_ADDRESS_INVALID)])

    subject = StringField(
        validators=[DataOptional()])

    body = HtmlTextAreaField(
        validators=[DataOptional()])

    attachments = FileField(
        validators=[DataOptional()],
        render_kw={'multiple': True})

    submit = SubmitField()

    def as_dict(self, attachment_encoder: AttachmentEncoder) -> dict:
        form = {key: value for (key, value) in self.data.items() if value}
        form.pop('submit', None)

        attachments = request.files.getlist(self.attachments.name)
        to = _split_emails(form.get('to'))
        cc = _split_emails(form.get('cc'))
        bcc = _split_emails(form.get('bcc'))

        sent_at = None
        if all(_is_local_message(address) for address in chain(to, cc, bcc)):
            sent_at = datetime.utcnow().strftime('%Y-%m-%d %H:%M')

        form['sent_at'] = sent_at
        form['read'] = True
        form['from'] = current_user.email
        form['to'] = to
        form['cc'] = cc
        form['bcc'] = bcc
        form['body'] = form.get('body')
        form['subject'] = form.get('subject', i8n.EMAIL_NO_SUBJECT)
        form['attachments'] = list(_attachments_as_dict(attachments,
                                                        attachment_encoder))
        return form

    def _populate(self, email: Optional[dict], to: Optional[str]):
        pass

    @classmethod
    def _new_instance_for(cls, action_name: Optional[str]):
        if not action_name:
            return cls(request.form)

        try:
            clazz = next(clazz for clazz in NewEmailForm.__subclasses__()
                         if getattr(clazz, 'action_name', None) == action_name)
        except StopIteration:
            return None
        else:
            return clazz(request.form)

    @classmethod
    def from_request(cls, email_store: EmailStore):
        action_name = request.args.get('action')
        form = cls._new_instance_for(action_name)
        if not form:
            return None

        to = request.args.get('to')
        uid = request.args.get('uid')
        reference = None
        if uid:
            reference = email_store.get(uid)
            if not current_user.can_access(reference):
                reference = None
        form._populate(reference, to)

        return form


class ToEmailForm(NewEmailForm):
    action_name = 'to'

    def _populate(self, email: Optional[dict], to: Optional[str]):
        self.to.data = to or ''


class ReplyEmailForm(NewEmailForm):
    action_name = 'reply'

    def _populate(self, email: Optional[dict], to: Optional[str]):
        if not email:
            return

        self.to.data = email.get('from', '')
        self.subject.data = 'Re: {}'.format(email.get('subject', ''))
        self.body.data = render_template('emails/reply.html', email=email)


class ReplyAllEmailForm(NewEmailForm):
    action_name = 'reply_all'

    def _populate(self, email: Optional[dict], to: Optional[str]):
        if not email:
            return

        self.to.data = _join_emails(email.get('from'), *email.get('cc', []))
        self.subject.data = 'Re: {}'.format(email.get('subject', ''))
        self.body.data = render_template('emails/reply.html', email=email)


class ForwardEmailForm(NewEmailForm):
    action_name = 'forward'

    def _populate(self, email: Optional[dict], to: Optional[str]):
        if not email:
            return

        self.subject.data = 'Fwd: {}'.format(email.get('subject', ''))
        self.body.data = render_template('emails/forward.html', email=email)


def _attachments_as_dict(
        filestorages: Iterable[FileStorage],
        attachment_encoder: AttachmentEncoder) -> Iterable[dict]:

    for filestorage in filestorages:
        filename = filestorage.filename
        content = attachment_encoder.encode(filestorage.stream.read())
        if filename and content:
            yield {'filename': filename, 'content': content}


def _is_local_message(address: str) -> bool:
    host = address.split('@')[-1]
    return host.lower() == AppConfig.CLIENT_EMAIL_HOST.lower()


def _join_emails(*emails: str) -> str:
    delimiter = '{0} '.format(AppConfig.EMAIL_ADDRESS_DELIMITER)
    return delimiter.join(email for email in emails if email)


def _split_emails(emails: Optional[str]) -> List[str]:
    if not emails:
        return []

    addresses = emails.split(AppConfig.EMAIL_ADDRESS_DELIMITER)
    return [address.strip() for address in addresses]



from datetime import timedelta
from functools import wraps

from flask import request
from flask_migrate import Migrate
from flask_security import LoginForm as _LoginForm
from flask_security import RegisterForm as _RegisterForm
from flask_security import RoleMixin
from flask_security import SQLAlchemyUserDatastore
from flask_security import Security
from flask_security import UserMixin
from flask_security import login_required as _login_required
from flask_security import roles_required
from flask_security.forms import email_required
from flask_security.forms import email_validator
from flask_security.forms import unique_user_email
from flask_sqlalchemy import SQLAlchemy
from sqlalchemy.exc import IntegrityError
from sqlalchemy.exc import OperationalError
from wtforms import IntegerField
from wtforms.validators import Regexp

from opwen_email_client.util.wtforms import SuffixedStringField
from opwen_email_client.webapp import app
from opwen_email_client.webapp.config import AppConfig
from opwen_email_client.webapp.config import i8n


_db = SQLAlchemy(app)

# noinspection PyUnresolvedReferences
_roles_users = _db.Table(
    'roles_users',
    _db.Column('user_id', _db.Integer(), _db.ForeignKey('user.id')),
    _db.Column('role_id', _db.Integer(), _db.ForeignKey('role.id')))


# noinspection PyUnresolvedReferences
class User(_db.Model, UserMixin):
    id = _db.Column(_db.Integer(), primary_key=True)
    email = _db.Column(_db.String(255), unique=True, index=True)
    password = _db.Column(_db.String(255), nullable=False)
    active = _db.Column(_db.Boolean(), default=True)
    last_login_at = _db.Column(_db.DateTime())
    current_login_at = _db.Column(_db.DateTime())
    last_login_ip = _db.Column(_db.String(128))
    current_login_ip = _db.Column(_db.String(128))
    login_count = _db.Column(_db.Integer())
    timezone_offset_minutes = _db.Column(_db.Integer(), nullable=False,
                                         default=0)
    language = _db.Column(_db.String(8))
    roles = _db.relationship('Role', secondary=_roles_users,
                             backref=_db.backref('users', lazy='dynamic'))

    @property
    def is_admin(self) -> bool:
        return self.has_role('admin')

    def format_last_login(self, timezone_offset_minutes) -> str:
        if not self.last_login_at:
            return ''

        date = self.last_login_at - timedelta(minutes=timezone_offset_minutes)
        return date.strftime('%Y-%m-%d %H:%M')

    def save(self):
        _db.session.add(self)
        _db.session.commit()

    def can_access(self, email: dict) -> bool:
        actors = set()
        actors.add(email.get('from'))
        actors.update(email.get('to', []))
        actors.update(email.get('cc', []))
        actors.update(email.get('bcc', []))

        return self.email in actors


# noinspection PyUnresolvedReferences
class Role(_db.Model, RoleMixin):
    id = _db.Column(_db.Integer(), primary_key=True)
    name = _db.Column(_db.String(32), unique=True)
    description = _db.Column(_db.String(255))


# noinspection PyClassHasNoInit
class LoginForm(_LoginForm):
    email = SuffixedStringField(
        suffix='@{}'.format(AppConfig.CLIENT_EMAIL_HOST))


# noinspection PyClassHasNoInit
class RegisterForm(_RegisterForm):
    email = SuffixedStringField(
        suffix='@{}'.format(AppConfig.CLIENT_EMAIL_HOST),
        validators=[Regexp('^[a-zA-Z0-9-.@]*$', message=i8n.EMAIL_CHARACTERS),
                    email_required, email_validator, unique_user_email])

    timezone_offset_minutes = IntegerField(default=0)


user_datastore = SQLAlchemyUserDatastore(_db, User, Role)

try:
    _db.create_all()
except OperationalError:
    pass

Migrate(app, _db)
Security(app, user_datastore, register_form=RegisterForm, login_form=LoginForm)

admin_role = 'admin'
user_datastore.find_or_create_role(name=admin_role)

try:
    user_datastore.commit()
except IntegrityError:
    user_datastore.db.session.rollback()


def login_required(func):
    if app.config.get('TESTING'):
        return func

    return _login_required(func)


def admin_required(func):
    if app.config.get('TESTING'):
        return func

    preshared_secret = app.config.get('ADMIN_SECRET')
    if preshared_secret:
        @wraps(func)
        def decorated_view(*args, **kwargs):
            secret = request.args.get('secret')
            if secret == preshared_secret:
                return func(*args, **kwargs)
            return roles_required(admin_role)(func)(*args, **kwargs)
        return decorated_view

    return roles_required(admin_role)(func)



from abc import ABCMeta
from abc import abstractmethod
from json import dumps
from json import loads
from typing import TypeVar

T = TypeVar('T')


class Serializer(metaclass=ABCMeta):
    @abstractmethod
    def serialize(self, obj: T) -> bytes:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def deserialize(self, serialized: bytes) -> T:
        raise NotImplementedError  # pragma: no cover


class JsonSerializer(Serializer):
    _encoding = 'utf-8'
    _separators = (',', ':')

    def serialize(self, obj):
        serialized = dumps(obj, separators=self._separators)
        return serialized.encode(self._encoding)

    def deserialize(self, serialized):
        decoded = serialized.decode(self._encoding)
        return loads(decoded)



from ast import literal_eval
from os import getenv as _getenv
from os import listdir
from os.path import isdir
from os.path import join
from typing import Iterable
from typing import TypeVar

T = TypeVar('T')


def getenv(key: str, default: T=None) -> T:
    value = _getenv(key, default)
    try:
        return literal_eval(value)
    except (ValueError, SyntaxError):
        return value


def subdirectories(root: str) -> Iterable[str]:
    try:
        return (sub for sub in listdir(root) if isdir(join(root, sub)))
    except OSError:
        return []



from flask import render_template

from opwen_email_client.domain.email.store import EmailStore
from opwen_email_client.domain.email.sync import Sync
from opwen_email_client.webapp.config import i8n


class SyncEmails(object):
    def __init__(self, email_store: EmailStore, email_sync: Sync):
        self._email_store = email_store
        self._email_sync = email_sync

    def _upload(self):
        pending = self._email_store.pending()
        uploaded = self._email_sync.upload(pending)
        self._email_store.mark_sent(uploaded)

    def _download(self):
        downloaded = self._email_sync.download()
        self._email_store.create(downloaded)

    def _sync(self):
        self._upload()
        self._download()

    def __call__(self):
        self._sync()


class SendWelcomeEmail(object):
    def __init__(self, to: str, time, email_store: EmailStore):
        self._to = to
        self._time = time
        self._email_store = email_store

    def __call__(self, *args, **kwargs):
        email_body = render_template('emails/account_finalized.html',
                                     email=self._to)
        self._email_store.create([{
            'sent_at': self._time.strftime("%Y-%m-%d %H:%M"),
            'to': [self._to],
            'from': 'info@ascoderu.ca',
            'subject': i8n.WELCOME,
            'body': email_body,
        }])



from opwen_email_client.webapp.ioc import create_app

app = create_app()

from opwen_email_client.webapp import login  # noqa: F401,E402
from opwen_email_client.webapp import views  # noqa: F401,E402



from abc import ABCMeta
from abc import abstractmethod
from os import getenv
from os import path
from typing import Tuple

from requests import Response
from requests import get as http_get
from requests import post as http_post


class EmailServerClient(metaclass=ABCMeta):
    @abstractmethod
    def upload(self, resource_id: str, container: str):
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def download(self) -> Tuple[str, str]:
        raise NotImplementedError  # pragma: no cover


class HttpEmailServerClient(EmailServerClient):
    _supported_resource_type = 'azure-blob'

    def __init__(self, read_api: str, write_api: str, client_id: str):
        self._read_api = read_api
        self._write_api = write_api
        self._client_id = client_id

    @property
    def _upload_url(self) -> str:
        return 'http://{host}/api/email/upload/{client_id}'.format(
            client_id=self._client_id,
            host=self._write_api)

    @property
    def _download_url(self) -> str:
        return 'http://{host}/api/email/download/{client_id}'.format(
            client_id=self._client_id,
            host=self._read_api)

    def upload(self, resource_id, container):
        payload = {
            'resource_id': resource_id,
            'container_name': container,
            'resource_type': self._supported_resource_type,
        }

        response = http_post(self._upload_url, json=payload)
        response.raise_for_status()

    def download(self):
        response = http_get(self._download_url)
        response.raise_for_status()

        resource_id, resource_container = self._validate(response)

        return resource_id, resource_container

    def _validate(self, response: Response):
        try:
            payload = response.json()
        except ValueError:
            payload = {}

        resource_id = payload.get('resource_id', '')
        resource_container = payload.get('resource_container', '')
        resource_type = payload.get('resource_type', '').lower()

        if resource_type and resource_type != self._supported_resource_type:
            raise ValueError('unsupported resource type: {}'
                             .format(resource_type))

        return resource_id, resource_container


class LocalEmailServerClient(EmailServerClient):
    def __init__(self, *args, **kwargs):
        pass

    def download(self) -> Tuple[str, str]:
        container, resource_id = 'to-lokole', 'emails.pack'
        local_file = path.join(getenv('AZURE_ROOT'), container, resource_id)
        if not path.isfile(local_file):
            return '', ''
        return resource_id, container

    def upload(self, resource_id: str, container: str):
        upload_directory = path.join(getenv('AZURE_ROOT'), container)
        print('Uploaded to {}'.format(upload_directory))



from setuptools import find_packages
from setuptools import setup


with open('requirements.txt') as fobj:
    install_requires = [line.strip() for line in fobj]


with open('README.rst') as fobj:
    long_description = fobj.read()


with open('version.txt') as fobj:
    version = fobj.read().strip()


packages = find_packages(exclude=['tests*'])
scripts = ['manage.py']


setup(
    name='opwen_email_client',
    version=version,
    author='Clemens Wolff',
    author_email='clemens.wolff+pypi@gmail.com',
    packages=packages,
    url='https://github.com/ascoderu/opwen-webapp',
    license='License :: OSI Approved :: Apache Software License',
    description='Email client for the Opwen project',
    long_description=long_description,
    include_package_data=True,
    scripts=scripts,
    install_requires=install_requires)



from itertools import islice
from typing import Iterable
from typing import TypeVar

T = TypeVar('T')


class Pagination(object):
    def __init__(self, items: Iterable[T], page: int, page_size: int):
        if page < 1:
            raise ValueError('page must be greater than or equal to 1')

        start = (page - 1) * page_size
        stop = page * page_size
        self._items = list(islice(items, start, stop))
        self.page = page
        self.page_size = page_size

    def __iter__(self):
        return iter(self._items)

    @property
    def has_prevpage(self) -> bool:
        return self.page != 1

    @property
    def has_nextpage(self) -> bool:
        return len(self._items) == self.page_size
